from __future__ import absolute_import

import time
import typing as tp
from struct import pack

# noinspection PyProtectedMember
from . import logger
from .errors import OutOfBoundsError
from .geometry import Region
from .utils import general_utils

if tp.TYPE_CHECKING:
    from .eyes_base import EyesBase
    from .utils.custom_types import (
        Num,
        RunningSession,
        AppOutput,
        UserInputs,
        MatchResult,
    )
    from .server_connector import ServerConnector
    from .match import ImageMatchSettings
    from .capture import EyesScreenshot

__all__ = ("MatchWindowTask",)


# TODO: remove Eyes and Target dependencies from here


class MatchWindowTask(object):
    """
    Handles matching of output with the expected output
    (including retry and 'ignore mismatch' when needed).
    """

    MATCH_INTERVAL = 0.5

    MINIMUM_MATCH_TIMEOUT = 60  # Milliseconds

    def __init__(self, eyes, server_connector, running_session, default_retry_timeout):
        # type: (EyesBase, ServerConnector, RunningSession, Num) -> None
        """
        Ctor.

        :param eyes: The Eyes instance which created this task.
        :param server_connector: The agent connector to use for communication.
        :param running_session:  The current eyes session.
        :param default_retry_timeout: The default match timeout. (milliseconds)
        """
        self._eyes = eyes
        self._server_connector = server_connector
        self._running_session = running_session
        self._default_retry_timeout = (
            default_retry_timeout / 1000.0
        )  # type: Num # since we want the time in seconds.
        self._last_screenshot = None  # type: tp.Optional[EyesScreenshot]

    @staticmethod
    def _create_match_data_bytes(
        app_output,  # type: AppOutput
        user_inputs,  # type: UserInputs
        tag,  # type: tp.Text
        ignore_mismatch,  # type: bool
        screenshot,  # type: EyesScreenshot
        default_match_settings,  # type: ImageMatchSettings
        target,
        ignore=None,  # type: tp.Optional[tp.List]
        floating=None,  # type: tp.Optional[tp.List]
    ):
        # type: (...) -> bytes
        if ignore is None:
            ignore = []
        if floating is None:
            floating = []

        match_data = {
            "IgnoreMismatch": ignore_mismatch,
            "Options": {
                "Name": tag,
                "UserInputs": user_inputs,
                "ImageMatchSettings": {
                    "MatchLevel": default_match_settings.match_level,
                    "IgnoreCaret": target.values.ignore_caret,
                    "Exact": default_match_settings.exact_settings,
                    "Ignore": ignore,
                    "Floating": floating,
                },
                "IgnoreMismatch": ignore_mismatch,
                "Trim": {"Enabled": False},
            },
            "UserInputs": user_inputs,
            "AppOutput": app_output,
            "tag": tag,
        }
        match_data_json_bytes = general_utils.to_json(match_data).encode(
            "utf-8"
        )  # type: bytes
        match_data_size_bytes = pack(">L", len(match_data_json_bytes))  # type: bytes
        screenshot_bytes = screenshot.bytes  # type: bytes
        body = match_data_size_bytes + match_data_json_bytes + screenshot_bytes
        return body

    @staticmethod
    def _get_dynamic_regions(target, eyes_screenshot):
        ignore = []  # type: tp.List[Region]
        floating = []  # type: tp.List[Region]
        if target is not None:
            for region_wrapper in target.values.ignore_regions:
                try:
                    current_region = region_wrapper.get_region(eyes_screenshot)
                    ignore.append(current_region)
                except OutOfBoundsError as err:
                    logger.info(
                        "WARNING: Region specified by {} is out of bounds! {}".format(
                            region_wrapper, err
                        )
                    )
            for floating_wrapper in target.values.floating_regions:
                try:
                    current_floating = floating_wrapper.get_region(eyes_screenshot)
                    floating.append(current_floating)
                except OutOfBoundsError as err:
                    logger.info(
                        "WARNING: Floating region specified by {} is out of bounds! {}".format(
                            floating_wrapper, err
                        )
                    )
        return {"ignore": ignore, "floating": floating}

    def _prepare_match_data_for_window(
        self,
        tag,  # type: tp.Text
        user_inputs,  # type: UserInputs
        default_match_settings,  # type: ImageMatchSettings
        target,
        ignore_mismatch=False,
    ):
        # type: (...) -> bytes
        title = self._eyes._title
        # TODO: Refactor this
        if hasattr(self._eyes, "_hide_scrollbars_if_needed"):
            with self._eyes._hide_scrollbars_if_needed():  # type: ignore
                self._last_screenshot = self._eyes.get_screenshot(
                    hide_scrollbars_called=True
                )
        else:
            self._last_screenshot = self._eyes.get_screenshot()

        dynamic_regions = MatchWindowTask._get_dynamic_regions(
            target, self._last_screenshot
        )
        app_output = {"title": title, "screenshot64": None}  # type: AppOutput

        if self._eyes.send_dom:
            dom_json = self._eyes._try_capture_dom()
            if dom_json:
                dom_url = self._eyes._try_post_dom_snapshot(dom_json)
                if dom_url is None:
                    logger.warning("Failed to upload DOM. Skipping...")
                else:
                    app_output["DomUrl"] = dom_url

        logger.debug("AppOutput: {}".format(app_output))
        return self._create_match_data_bytes(
            app_output,
            user_inputs,
            tag,
            ignore_mismatch,
            self._last_screenshot,
            default_match_settings,
            target,
            dynamic_regions["ignore"],
            dynamic_regions["floating"],
        )

    def _run_with_intervals(self, prepare_match_data_options, retry_timeout):
        # type: (tp.Dict, Num) -> MatchResult
        """
        Includes retries in case the screenshot does not match.
        """
        logger.debug("Matching with intervals...")
        # We intentionally take the first screenshot before starting the timer,
        # to allow the page just a tad more time to stabilize.
        prepare_match_data_options = prepare_match_data_options.copy()
        prepare_match_data_options["ignore_mismatch"] = True
        data = self._prepare_match_data_for_window(**prepare_match_data_options)

        # Start the timer.
        start = time.time()
        logger.debug("First match attempt...")
        as_expected = self._server_connector.match_window(self._running_session, data)
        if as_expected:
            return {"as_expected": True, "screenshot": self._last_screenshot}
        retry = time.time() - start
        logger.debug("Failed. Elapsed time: {0:.1f} seconds".format(retry))

        while retry < retry_timeout:
            logger.debug("Matching...")
            time.sleep(self.MATCH_INTERVAL)

            data = self._prepare_match_data_for_window(**prepare_match_data_options)
            as_expected = self._server_connector.match_window(
                self._running_session, data
            )
            if as_expected:
                return {"as_expected": True, "screenshot": self._last_screenshot}
            retry = time.time() - start
            logger.debug("Elapsed time: {0:.1f} seconds".format(retry))

        # One last try
        logger.debug("One last matching attempt...")
        prepare_match_data_options["ignore_mismatch"] = False
        data = self._prepare_match_data_for_window(**prepare_match_data_options)
        as_expected = self._server_connector.match_window(self._running_session, data)
        return {"as_expected": as_expected, "screenshot": self._last_screenshot}

    def match_window(
        self,
        retry_timeout,  # type: Num
        tag,  # type: str
        user_inputs,  # UserInputs
        default_match_settings,  # type: ImageMatchSettings
        target,
        ignore_mismatch,
        run_once_after_wait=False,
    ):
        # type: (...) -> MatchResult
        """
        Performs a match for the window.

        :param retry_timeout: Amount of time until it retries.
        :param tag: The name of the tag (optional).
        :param user_inputs: The user input.
        :param default_match_settings: The default match settings for the session.
        :param target: The target of the check_window call.
        :param ignore_mismatch: True if the server should ignore a negative result
                                for the visual validation.
        :param run_once_after_wait: Whether or not to run again after waiting.
        :return: The result of the run.
        """
        prepare_match_data_options = dict(
            tag=tag,
            user_inputs=user_inputs,
            default_match_settings=default_match_settings,
            target=target,
            ignore_mismatch=ignore_mismatch,
        )

        if 0 < retry_timeout < MatchWindowTask.MINIMUM_MATCH_TIMEOUT:
            raise ValueError(
                "Match timeout must be at least 60ms, got {} instead.".format(
                    retry_timeout
                )
            )
        if retry_timeout < 0:
            retry_timeout = self._default_retry_timeout
        else:
            retry_timeout /= 1000.0

        logger.debug("Match timeout set to: {0} seconds".format(retry_timeout))
        start = time.time()
        if run_once_after_wait or retry_timeout == 0:
            logger.debug("Matching once...")
            # If the load time is 0, the sleep would immediately return anyway.
            time.sleep(retry_timeout)
            data = self._prepare_match_data_for_window(**prepare_match_data_options)
            as_expected = self._server_connector.match_window(
                self._running_session, data
            )
            result = {
                "as_expected": as_expected,
                "screenshot": self._last_screenshot,
            }  # type: MatchResult
        else:
            result = self._run_with_intervals(prepare_match_data_options, retry_timeout)

        logger.debug("Match result: {0}".format(result["as_expected"]))
        elapsed_time = time.time() - start
        logger.debug("_run(): Completed in {0:.1f} seconds".format(elapsed_time))
        return result



from __future__ import absolute_import

import os
import time
import typing as tp

import requests
from requests.packages import urllib3

from . import logger
from .test_results import TestResults
from .utils import general_utils
from .utils.compat import gzip_compress, urljoin  # type: ignore

if tp.TYPE_CHECKING:
    from .utils.custom_types import RunningSession, SessionStartInfo, Num

# Prints out all data sent/received through 'requests'
# import httplib
# httplib.HTTPConnection.debuglevel = 1

# Remove Unverified SSL warnings propagated by requests' internal urllib3 module
if hasattr(urllib3, "disable_warnings") and callable(urllib3.disable_warnings):
    urllib3.disable_warnings()

__all__ = ("ServerConnector",)


class _RequestsCommunicator(object):
    LONG_REQUEST_DELAY = 2  # seconds
    MAX_LONG_REQUEST_DELAY = 10  # seconds
    LONG_REQUEST_DELAY_MULTIPLICATIVE_INCREASE_FACTOR = 1.5

    api_key = None
    timeout = None
    server_url = None  # type: tp.Optional[tp.Text]

    def __init__(self, timeout, headers, api_session_running):
        # type: (int, tp.Dict, tp.Text) -> None
        self.timeout = timeout
        self.headers = headers.copy()
        self.api_session_running = api_session_running

    @property
    def endpoint_uri(self):
        return urljoin(self.server_url, self.api_session_running)

    def _request(self, method, url_resource, **kwargs):
        if url_resource is not None:
            # makes URL relative
            url_resource = url_resource.lstrip("/")
        url_resource = urljoin(self.endpoint_uri, url_resource)

        params = dict(apiKey=self.api_key)
        params.update(kwargs.get("params", {}))
        headers = kwargs.get("headers", self.headers)
        timeout = kwargs.get("timeout", self.timeout)

        response = method(
            url_resource,
            data=kwargs.get("data", None),
            verify=False,
            params=params,
            headers=headers,
            timeout=timeout,
        )
        response.raise_for_status()
        return response

    def _long_request(self, method, url_resource, **kwargs):
        headers = kwargs["headers"].copy()
        headers["Eyes-Expect"] = "202-accepted"
        for delay in self.request_delay():
            # Sending the current time of the request (in RFC 1123 format)
            headers["Eyes-Date"] = time.strftime(
                "%a, %d %b %Y %H:%M:%S GMT", time.gmtime()
            )
            kwargs["headers"] = headers
            response = self._request(method, url_resource, **kwargs)
            if response.status_code != 202:
                return response
            logger.debug("Still running... Retrying in {}s".format(delay))
        else:
            raise requests.Timeout("Couldn't process request")

    def post(self, url_resource=None, **kwargs):
        return self._request(requests.post, url_resource, **kwargs)

    def long_delete(self, url_resource=None, **kwargs):
        return self._long_request(requests.delete, url_resource, **kwargs)

    @staticmethod
    def request_delay(
        first_delay=LONG_REQUEST_DELAY,
        step_factor=LONG_REQUEST_DELAY_MULTIPLICATIVE_INCREASE_FACTOR,
        max_delay=MAX_LONG_REQUEST_DELAY,
    ):
        delay = _RequestsCommunicator.LONG_REQUEST_DELAY  # type: Num
        while True:
            yield delay
            time.sleep(first_delay)
            delay = delay * step_factor
            if delay > max_delay:
                raise StopIteration


class ServerConnector(object):
    """
    Provides an API for communication with the Applitools server.
    """

    DEFAULT_TIMEOUT = 60 * 5  # Seconds
    DEFAULT_HEADERS = {"Accept": "application/json", "Content-Type": "application/json"}
    DEFAULT_SERVER_URL = "https://eyesapi.applitools.com"
    API_SESSIONS_RUNNING = "/api/sessions/running/"

    def __init__(self, server_url):
        # type: (tp.Optional[tp.Text]) -> None
        """
        Ctor.

        :param server_url: The url of the Applitools server.
        """
        self._request = _RequestsCommunicator(
            timeout=ServerConnector.DEFAULT_TIMEOUT,
            headers=ServerConnector.DEFAULT_HEADERS,
            api_session_running=ServerConnector.API_SESSIONS_RUNNING,
        )
        self.server_url = server_url

    @property
    def server_url(self):
        # type: () -> tp.Text
        return self._request.server_url

    @server_url.setter
    def server_url(self, server_url):
        # type: (tp.Text) -> None
        if server_url is None:
            self._request.server_url = self.DEFAULT_SERVER_URL
        else:
            self._request.server_url = server_url

    @property
    def api_key(self):
        if self._request.api_key is None:
            # if api_key is None the error will be raised in EyesBase._open_base
            self._request.api_key = os.environ.get("APPLITOOLS_API_KEY", None)
        return self._request.api_key

    @api_key.setter
    def api_key(self, api_key):
        self._request.api_key = api_key

    @property
    def timeout(self):
        if self._request.timeout is None:
            self._request.timeout = self.DEFAULT_TIMEOUT
        return self._request.timeout

    @timeout.setter
    def timeout(self, timeout):
        self._request.timeout = timeout

    # TODO: Add Proxy
    def start_session(self, session_start_info):
        # type: (SessionStartInfo) -> RunningSession
        """
        Starts a new running session in the agent. Based on the given parameters,
        this running session will either be linked to an existing session, or to
        a completely new session.

        :param session_start_info: The start params for the session.
        :return: Represents the current running session.
        """
        data = '{"startInfo": %s}' % (general_utils.to_json(session_start_info))
        response = self._request.post(data=data)
        parsed_response = response.json()
        return dict(
            session_id=parsed_response["id"],
            session_url=parsed_response["url"],
            is_new_session=(response.status_code == requests.codes.created),
        )

    def stop_session(self, running_session, is_aborted, save):
        # type: (RunningSession, bool, bool) -> TestResults
        """
        Stops a running session in the Eyes server.

        :param running_session: The session to stop.
        :param is_aborted: Whether the server should mark this session as aborted.
        :param save: Whether the session should be automatically saved if it is not aborted.
        :return: Test results of the stopped session.
        """
        logger.debug("Stop session called..")
        params = {"aborted": is_aborted, "updateBaseline": save, "apiKey": self.api_key}
        headers = ServerConnector.DEFAULT_HEADERS.copy()
        response = self._request.long_delete(
            url_resource=running_session["session_id"], params=params, headers=headers
        )
        pr = response.json()
        logger.debug("stop_session(): parsed response: {}".format(pr))
        return TestResults(
            pr["steps"],
            pr["matches"],
            pr["mismatches"],
            pr["missing"],
            pr["exactMatches"],
            pr["strictMatches"],
            pr["contentMatches"],
            pr["layoutMatches"],
            pr["noneMatches"],
            pr["status"],
        )

    def match_window(self, running_session, data):
        # type: (RunningSession, bytes) -> bool
        """
        Matches the current window to the immediate expected window in the Eyes server.
        Notice that a window might be matched later at the end of the test, even if it
        was not immediately matched in this call.

        :param running_session: The current session that is running.
        :param data: The data for the requests.post.
        :return: The parsed response.
        """
        # logger.debug("Data length: %d, data: %s" % (len(data), repr(data)))
        # Using the default headers, but modifying the "content type" to binary
        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = "application/octet-stream"
        response = self._request.post(
            url_resource=running_session["session_id"], data=data, headers=headers
        )
        return response.json()["asExpected"]

    def post_dom_snapshot(self, dom_json):
        # type: (tp.Text) -> tp.Optional[tp.Text]
        """
        Upload the DOM of the tested page.
        Return an URL of uploaded resource which should be posted to :py:   `AppOutput`.
        """
        headers = ServerConnector.DEFAULT_HEADERS.copy()
        headers["Content-Type"] = "application/octet-stream"
        dom_bytes = gzip_compress(dom_json.encode("utf-8"))
        response = self._request.post(
            url_resource="data", data=dom_bytes, headers=headers
        )
        dom_url = None
        if response.ok:
            dom_url = response.headers["Location"]
        return dom_url



from __future__ import absolute_import

import math
import typing as tp
from collections import OrderedDict

from .utils import argument_guard
from .metadata import CoordinatesType
from .errors import EyesError

if tp.TYPE_CHECKING:
    from applitools.core.utils.custom_types import ViewPort

__all__ = ("Point", "Region")


class DictAccessMixin(object):
    def __getitem__(self, item):
        if item not in self.__slots__:
            raise KeyError
        return getattr(self, item)


class StateMixin(object):
    def __getstate__(self):
        return OrderedDict([(name, getattr(self, name)) for name in self.__slots__])

    # Required is required in order for jsonpickle to work on this object.
    # noinspection PyMethodMayBeStatic
    def __setstate__(self, state):
        raise EyesError("Cannot create Point instance from dict!")


class Point(DictAccessMixin, StateMixin):
    """
    A point with the coordinates (x,y).
    """

    __slots__ = ("x", "y")

    def __init__(self, x=0, y=0):
        # type: (float, float) -> None
        self.x = int(round(x))
        self.y = int(round(y))

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __iadd__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, scalar):
        return Point(self.x * scalar, self.y * scalar)

    def __div__(self, scalar):
        return Point(self.x / scalar, self.y / scalar)

    def __str__(self):
        return "({0}, {1})".format(self.x, self.y)

    def __bool__(self):
        return self.x and self.y

    @classmethod
    def create_top_left(cls):
        return cls(0, 0)

    def length(self):
        # type: () -> float
        """
        Returns the distance from (0, 0).
        """
        return math.sqrt(self.x ** 2 + self.y ** 2)

    def distance_to(self, p):
        # type: (Point) -> float
        """
        Calculate the distance between two points.

        :return: The distance to p.
        """
        return (self - p).length()

    def as_tuple(self):
        # type: () -> tuple
        """
        Return the point as a tuple.

        :return: Point as tuple.
        """
        return self.x, self.y

    def clone(self):
        # type: () -> Point
        """
        Return a full copy of this point.

        :return: Cloned point.
        """
        return Point(self.x, self.y)

    def move_to(self, x, y):
        # type: (int, int) -> None
        """
        Moves the point to new x, y.

        :param x: Coordinate x.
        :param y: Coordinate y.
        """
        self.x = x
        self.y = y

    def offset(self, dx, dy):
        # type: (int, int) -> Point
        """
        Move to new (x+dx,y+dy).

        :param dx: Offset to move coordinate x.
        :param dy: Offset to move coordinate y.
        """
        self.x = self.x + dx
        self.y = self.y + dy
        return self

    def offset_by_location(self, location):
        # type: (Point) -> Point
        self.offset(location.x, location.y)
        return self

    def offset_negative(self, dx, dy):
        # type: (int, int) -> Point
        self.x -= dx
        self.y -= dy
        return self

    def rotate(self, rad):
        # type: (int) -> Point
        """
        Rotate counter-clockwise around the origin by rad radians.

        Positive y goes *up,* as in traditional mathematics.

        Interestingly, you can use this in y-down computer graphics, if
        you just remember that it turns clockwise, rather than
        counter-clockwise.

        :param rad: The radians to rotate the point.
        :return: The new position is returned as a new Point.
        """
        s, c = [f(rad) for f in (math.sin, math.cos)]
        x, y = (c * self.x - s * self.y, s * self.x + c * self.y)
        return Point(x, y)

    def rotate_about(self, p, theta):
        # type: (Point, int) -> Point
        """
        Rotate counter-clockwise around a point, by theta degrees.

        Positive y goes *up,* as in traditional mathematics.

        The new position is returned as a new Point.

        :param p: A point to rotate around.
        :param theta: Theta degrees to rotate around.
        :return: The result of the rotation.
        """
        result = self.clone()
        result.offset(-p.x, -p.y)
        result.rotate(theta)
        result.offset(p.x, p.y)
        return result

    def scale(self, scale_ratio):
        return Point(
            int(math.ceil(self.x * scale_ratio)), int(math.ceil(self.y * scale_ratio))
        )


class Region(DictAccessMixin, StateMixin):
    """
    A rectangle identified by left,top, width, height.
    """

    __slots__ = ("left", "top", "width", "height", "coordinates_type")

    def __init__(
        self,
        left=0,
        top=0,
        width=0,
        height=0,
        coordinates_type=CoordinatesType.SCREENSHOT_AS_IS,
    ):
        # type: (float, float, float, float, tp.Text) -> None
        self.left = int(round(left))
        self.top = int(round(top))
        self.width = int(round(width))
        self.height = int(round(height))
        self.coordinates_type = coordinates_type

    @classmethod
    def create_empty_region(cls):
        return cls(0, 0, 0, 0)

    @classmethod
    def from_region(cls, region):
        return cls(
            region.left,
            region.top,
            region.width,
            region.height,
            region.coordinates_type,
        )

    @classmethod
    def from_location_size(cls, location, size):
        return cls(location.x, location.y, size["width"], size["height"])

    @property
    def x(self):
        return self.left

    @property
    def y(self):
        return self.top

    @property
    def right(self):
        # type: () -> int
        return self.left + self.width

    @property
    def bottom(self):
        # type: () -> int
        return self.top + self.height

    @property
    def location(self):
        # type: () -> Point
        """Return the top-left corner as a Point."""
        return Point(self.left, self.top)

    @location.setter
    def location(self, point):
        # type: (Point) -> None
        """Sets the top left corner of the region"""
        argument_guard.not_none(point)
        self.left, self.top = point.x, point.y

    @property
    def bottom_right(self):
        # type: () -> Point
        """Return the bottom-right corner as a Point."""
        return Point(self.right, self.bottom)

    @property
    def size(self):
        # type: () -> ViewPort
        """
        :return: The size of the region.
        """
        return dict(width=self.width, height=self.height)

    def clone(self):
        # type: () -> Region
        """
        Clone the rectangle.

        :return: The new rectangle object.
        """
        return Region(self.left, self.top, self.width, self.height)

    def is_same(self, other):
        # type: (Region) -> bool
        """
        Checks whether the other rectangle has the same coordinates.

        :param other: The other rectangle to check with.
        :return: Whether or not the rectangles have same coordinates.
        :rtype: bool
        """
        return (
            self.left == other.left
            and self.top == other.top
            and self.width == other.width
            and self.height == other.height
        )

    def is_same_size(self, other):
        # type: (Region) -> bool
        """
        Checks whether the other rectangle is the same size.

        :param other: The other rectangle to check with.
        :return: Whether or not the rectangles are the same size.
        """
        return self.width == other.width and self.height == other.height

    def make_empty(self):
        """
        Sets the current instance as an empty instance
        """
        self.left = self.top = self.width = self.height = 0

    def clip_negative_location(self):
        """
        Sets the left/top values to 0 if the value is negative
        """
        self.left = max(self.left, 0)
        self.top = max(self.top, 0)

    @property
    def is_size_empty(self):
        # type: () -> bool
        """
        :return: true if the region's size is 0, false otherwise.
        """
        return self.width <= 0 or self.height <= 0

    @property
    def is_empty(self):
        # type: () -> bool
        """
        Checks whether the rectangle is empty.

        :return: True if the rectangle is empty. Otherwise False.
        """
        return self.left == self.top == self.width == self.height == 0

    def contains(self, pt):
        # type: (Point) -> bool
        """
        Return true if a point is inside the rectangle.

        :return: True if the point is inside the rectangle. Otherwise False.
        """
        x, y = pt.as_tuple()
        return self.left <= x <= self.right and self.top <= y <= self.bottom  # noqa

    def overlaps(self, other):
        # type: (Region) -> bool
        """
        Return true if a rectangle overlaps this rectangle.
        """
        return (
            self.left <= other.left <= self.right
            or other.left <= self.left <= other.right
        ) and (
            self.top <= other.top <= self.bottom
            or other.top <= self.top <= other.bottom
        )

    def intersect(self, other):
        # type: (Region) -> None
        # If the regions don't overlap, the intersection is empty
        if not self.overlaps(other):
            self.make_empty()
            return
        intersection_left = self.left if self.left >= other.left else other.left
        intersection_top = self.top if self.top >= other.top else other.top
        intersection_right = self.right if self.right <= other.right else other.right
        intersection_bottom = (
            self.bottom if self.bottom <= other.bottom else other.bottom
        )
        self.left, self.top = intersection_left, intersection_top
        self.width = intersection_right - intersection_left
        self.height = intersection_bottom - intersection_top

    def get_sub_regions(self, max_sub_region_size):
        # type: (tp.Dict) -> tp.List[Region]
        """
        Returns a list of Region objects which compose the current region.
        """
        sub_regions = []
        current_top = self.top
        while current_top < self.height:

            current_bottom = current_top + max_sub_region_size["height"]
            if current_bottom > self.height:
                current_bottom = self.height

            current_left = self.left
            while current_left < self.width:
                current_right = current_left + max_sub_region_size["width"]
                if current_right > self.width:
                    current_right = self.width

                current_height = current_bottom - current_top
                current_width = current_right - current_left

                sub_regions.append(
                    Region(current_left, current_top, current_width, current_height)
                )

                current_left += max_sub_region_size["width"]

            current_top += max_sub_region_size["height"]

        return sub_regions

    @property
    def middle_offset(self):
        # type: () -> Point
        return Point(int(round(self.width / 2)), int(round(self.height / 2)))

    def offset(self, dx, dy):
        location = self.location.offset(dx, dy)
        return Region(
            left=location.x,
            top=location.y,
            width=self.size["width"],
            height=self.size["height"],
        )

    def scale(self, scale_ratio):
        return Region(
            left=int(math.ceil(self.left * scale_ratio)),
            top=int(math.ceil(self.top * scale_ratio)),
            width=int(math.ceil(self.width * scale_ratio)),
            height=int(math.ceil(self.height * scale_ratio)),
        )

    def __str__(self):
        return "(%s, %s) %s x %s" % (self.left, self.top, self.width, self.height)



from . import image_utils, argument_guard  # noqa
from .general_utils import cached_property  # noqa
from .compat import ABC, range, iteritems  # noqa

__all__ = compat.__all__ + ("image_utils", "argument_guard")  # noqa



from .position_provider import (  # noqa
    PositionProvider,
    InvalidPositionProvider,
    PositionMemento,
)
from .region_provider import (  # noqa
    RegionProvider,
    NullRegionProvider,
    NULL_REGION_INSTANCE,
)

__all__ = position_provider.__all__ + region_provider.__all__  # noqa



from .triggers import *  # noqa
from .test_results import *  # noqa
from .match_window_task import *  # noqa
from .logger import *  # noqa
from .errors import *  # noqa
from .capture import *  # noqa
from .scaling import *  # noqa
from .eyes_base import *  # noqa
from .geometry import *  # noqa
from .server_connector import ServerConnector  # noqa
from .positioning import *  # noqa
from .match import *  # noqa
from .metadata import *  # noqa

__all__ = (
    triggers.__all__  # noqa
    + test_results.__all__  # noqa
    + match_window_task.__all__  # noqa
    + logger.__all__  # noqa
    + errors.__all__  # noqa
    + scaling.__all__  # noqa
    + capture.__all__  # noqa
    + eyes_base.__all__  # noqa
    + geometry.__all__  # noqa
    + positioning.__all__  # noqa
    + match.__all__  # noqa
    + metadata.__all__  # noqa
    + ("logger", "ServerConnector")  # noqa
)
